// Copyright (C) 2007-2021  CEA/DEN, EDF R&D, OPEN CASCADE
//
// Copyright (C) 2003-2007  OPEN CASCADE, EADS/CCR, LIP6, CEA/DEN,
// CEDRAT, EDF R&D, LEG, PRINCIPIA R&D, BUREAU VERITAS
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
//
// See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
//

//  SALOME Utils : general SALOME's definitions and tools
//  File   : Utils_SALOME_Exception.hxx
//  Author : Antoine YESSAYAN, EDF
//  Module : SALOME
//  $Header$
//
#pragma once

#include <exception>
#include <sstream>
#include <string>

#ifdef LOCALIZED
#undef LOCALIZED
#endif
#if defined(_DEBUG_) || defined(_DEBUG)
#define LOCALIZED(message) #message, __FILE__, __LINE__
#else
#define LOCALIZED(message) #message
#endif

//swig tool on Linux doesn't pass defines from header SALOME_Utils.hxx
//therefore (temporary solution) defines are placed below

#ifdef WIN32
#if defined UTILS_EXPORTS || defined OpUtil_EXPORTS
#define UTILS_EXPORT __declspec(dllexport)
#else
#define UTILS_EXPORT __declspec(dllimport)
#undef LOCALIZED
#define LOCALIZED(message) #message
#endif
#else
#define UTILS_EXPORT
#endif

class SALOME_Exception;

UTILS_EXPORT std::ostream &operator<<(std::ostream &, const SALOME_Exception &);

UTILS_EXPORT std::string makeText(const char *text, const char *fileName, const unsigned int lineNumber);

class UTILS_EXPORT SALOME_Exception : public std::exception
{
protected:
  std::string _text;
public:
  SALOME_Exception() = delete;
  SALOME_Exception(const std::string& text):_text(text) { }
  SALOME_Exception(const char *text, const char *fileName = nullptr, const unsigned int lineNumber = 0);
  virtual ~SALOME_Exception() noexcept;
  UTILS_EXPORT friend std::ostream &operator<<(std::ostream &os, const SALOME_Exception &ex);
  virtual const char *what(void) const noexcept;
};

#define THROW_SALOME_EXCEPTION(text)      \
{                                         \
    std::ostringstream oss; oss << text;  \
    throw SALOME_Exception(oss.str());    \
}
